.dataTable td, th {
    padding-top: 5px !important;
    padding-bottom: 5px !important;
    padding-left: 5px !important;
    padding-right: 8px !important;
}

.dataTable thead th {
    background-color: #9BC3B7 !important;;
    color: #404040 !important;;
    padding-right: 28px !important;
}

.dataTable tr:nth-child(odd) {
    background-color: #ededed !important;
}

.dataTable tbody tr.selected>*{
    color: #fcfcfc !important;
    background-color: #26669a7a;
    box-shadow: none !important;
}

.dataTable td:hover {
    color: #fcfcfc !important;
    background-color: #597a8d !important;
}

.dataTable tbody td.focus {
    outline: 2px solid #005eab !important;
}

.dt-button, .paginate_button, .dtsb-button {
    background-color: #f3f6f6 !important;
    background: none !important;
    color: #404040 !important;
    display: inline-block !important;
    border-radius: 2px !important;
    line-height: normal !important;
    white-space: nowrap !important;
    text-align: center !important;
    cursor: pointer !important;
    font-size: 100% !important;
    padding: 6px 12px 8px !important;
    border: 1px solid rgba(0, 0, 0, .1) !important;
    text-decoration: none !important;
    font-weight: 400 !important;
    font-family: Lato, proxima-nova, Helvetica Neue, Arial, sans-serif !important;
    box-shadow: inset 0 1px 2px -1px hsl(0deg 0% 100% / 50%), inset 0 -2px 0 0 rgb(0 0 0 / 10%) !important;
    outline-none: false !important;
    vertical-align: middle !important;
    *display: inline !important;
    zoom: 1 !important;
    -webkit-user-drag: none !important;
    -webkit-user-select: none !important;
    -moz-user-select: none !important;
    -ms-user-select: none !important;
    user-select: none !important;
    -webkit-transition: all .1s linear !important;
    -moz-transition: all .1s linear !important;
    transition: all .1s linear !important;
}

.dt-button:hover {
    background-color: #e5ebeb !important;
    color: #404040;
}

.dataTables_length {
    margin-left: 20px;
}

.dataTables_length select {
    background-color: #f3f6f6 !important;
    background: none !important;
    color: #404040 !important;
    display: inline-block !important;
    border-radius: 2px !important;
    line-height: normal !important;
    white-space: nowrap !important;
    text-align: center !important;
    cursor: pointer !important;
    font-size: 100% !important;
    padding: 6px 12px 8px !important;
    border: 1px solid rgba(0, 0, 0, .1) !important;
    text-decoration: none !important;
    font-weight: 400 !important;
    font-family: Lato, proxima-nova, Helvetica Neue, Arial, sans-serif !important;
    box-shadow: inset 0 1px 2px -1px hsl(0deg 0% 100% / 50%), inset 0 -2px 0 0 rgb(0 0 0 / 10%) !important;
    outline-none: false !important;
    vertical-align: middle !important;
    *display: inline !important;
    zoom: 1 !important;
    -webkit-user-drag: none !important;
    -webkit-user-select: none !important;
    -moz-user-select: none !important;
    -ms-user-select: none !important;
    user-select: none !important;
    -webkit-transition: all .1s linear !important;
    -moz-transition: all .1s linear !important;
    transition: all .1s linear !important;
}
